/*
 *	A virtual machine a.k.a. Robot Karel
 *
 *	(c) 2008 Anicka Bernathova <anicka@anicka.net>
 *	(c) 2008 Martin Mares <mj@ucw.cz>
 *
 *	This program can be freely distributed and used according to the
 *	terms of the GNU General Public License version 2.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>

typedef uint8_t byte;
typedef uint16_t u16;
typedef uint32_t u32;

#define X 46
#define Y 22
#define MAX_BYTES 65536
#define MAX_COMMANDS 1024
#define MAX_STEPS 65536
#undef DEBUG
#define IOBUF_SIZE 4096

/*
 *  bit 0 = wall up
 *  bit 1 = wall left
 *  bit 2 = wall down
 *  bit 3 = wall right
 *  bit 4 = help available here
 *  bit 5 = visited
 *  bit 6 = collision up
 *  bit 7 = collision left
 *  bit 8 = collision down
 *  bit 9 = collision right
 */

static u16 maze[Y][X] = {
#include "maze.h"
};

static int delta[4][2] = { {0,-1}, {-1,0}, {0,1}, {1,0} };

enum cmd {
	CMD_END,
	CMD_RIGHT,
	CMD_LEFT,
	CMD_STEP,
	CMD_IS_WALL,
	CMD_IS_NOT_WALL,
	CMD_IS_HOME,
	CMD_IS_NOT_HOME,
	CMD_RESTART,
};

static byte program[MAX_COMMANDS+2];

static byte source[MAX_BYTES];
static int source_len;

static int cgi_mode;
static int end_x, end_y, end_dir;

static int x, y;

static struct {
	u32 hash;
	int cmd;
} cmd_hash[] = {
#include "cmds.h"
};

static void html_head(void);
static void html_tail(void);
static void echo(char *c);

static void __attribute__((noreturn)) error(char *msg)
{
	if (cgi_mode) {
		html_head();
		echo("<p>");
		echo(msg);
		echo("</p>\n");
		html_tail();
	} else {
		fprintf(stderr, "ERROR: %s\n", msg);
	}
	exit(0);
}

static void parse(void)
{
	byte *buf = source;
	unsigned proglen = 0;

	while (*buf) {
		unsigned c = *buf++;
		if (!(c >= 'a' && c <= 'z' || c >= 'A' && c <= 'Z'))
			continue;

		u32 hash = c;
		while (*buf) {
			c = *buf;
			if (!(c >= 'a' && c <= 'z' || c >= 'A' && c <= 'Z'))
				break;
			hash = hash*23 + c;
			buf++;
		}

		unsigned pos = hash % HASH_SIZE;
		while (cmd_hash[pos].cmd >= 0) {
			if (cmd_hash[pos].hash == hash) {
				if (proglen >= MAX_COMMANDS)
					error("Program obsahuje příliš mnoho příkazů. Takhle dlouhé programy moje paměť nezvládá.");
				program[proglen++] = cmd_hash[pos].cmd;
				break;
			}
			pos = (pos+1) % HASH_SIZE;
		}
	}
	program[proglen++] = CMD_END;	// Might be discarded by condition
	program[proglen++] = CMD_END;	// ... so Karel needs two end instructions to halt :)

#ifdef DEBUG
	printf("Parsed program:\n");
	for (unsigned i=0; i<proglen; i++)
		printf("%3d: %d\n", i, program[i]);
#endif
}

static void dump(void)
{
	for (int y=0; y<Y; y++) {
		for (int x=0; x<X; x++) {
			if (maze[y][x] & 64) printf("XXX");
			else if (maze[y][x] & 1) printf("---");
			else printf("...");
		}
		putchar('\n');
		for (int x=0; x<X; x++) {
			if (maze[y][x] & 128) putchar('X');
			else if (maze[y][x] & 2) putchar('|');
			else putchar('.');

			if (y == end_y && x == end_x) putchar("^<v>"[end_dir]);
			else if (maze[y][x] & 32) putchar('@');
			else if (maze[y][x] & 16) putchar('$');
			else putchar(' ');

			if (maze[y][x] & 512) putchar('X');
			else if (maze[y][x] & 8) putchar('|');
			else putchar('.');
		}
		putchar('\n');
		for (int x=0; x<X; x++) {
			if (maze[y][x] & 256) printf("XXX");
			else if (maze[y][x] & 4) printf("---");
			else printf("...");
		}
		putchar('\n');
	}
}

static void go(void)
{
	int step = 0;
	int i;
	int direction = 3;
	int cond = 1;

	maze[y][x] |= 1 << 5;
	for (i = 0; ; i++){
		if (step > MAX_STEPS)
			error("Program běžel příliš dlouho a byl násilně ukončen.");
		step++;
		switch (program[i]){
			case CMD_END:
				if (cond) {
					end_x = x;
					end_y = y;
					end_dir = direction;
					return;
				}
				cond = 1;
				break;
			case CMD_RIGHT:
				if (cond)
					direction = (direction + 3) % 4;
				cond = 1;
				break;	
			case CMD_LEFT:
				if (cond)
					direction = (direction + 1) % 4;
				cond = 1;
				break;
			case CMD_RESTART:
				if (cond)
					i = -1;
				cond = 1;
				break;
			case CMD_STEP:
				if (!cond){
					cond = 1;
					continue;
				}
				if (maze[y][x] & (1 << direction)){
					maze[y][x] |= 1 << (direction + 6);
					continue;
				}
				x += delta[direction][0];
				y += delta[direction][1];
				maze[y][x] |= 1 << 5;
				break;
			case CMD_IS_WALL:
				if (! (maze[y][x] & (1 << direction)))
					cond = 0;
				break;	
			case CMD_IS_NOT_WALL:
				if (maze[y][x] & (1 << direction))
					cond = 0;
				break;
			case CMD_IS_HOME:
				if (x || y)
					cond = 0;
				break;	
			case CMD_IS_NOT_HOME:
				if (!(x || y))
					cond = 0;
				break;	
		}		
	}
}

static void stdin_read(void)
{
	for (;;) {
		int rem = MAX_BYTES - source_len;
		if (!rem)
			error("Program je příliš dlouhý");
		int got = read(0, source + source_len, rem);
		if (got < 0)
			error("Chyba při čtení");
		if (!got)
			break;
		source_len += got;
	}
	source[source_len] = 0;
}

static byte iobuf[IOBUF_SIZE];
static byte *ioptr = iobuf;
static byte *ioend = iobuf + IOBUF_SIZE;

static void ioflush(void)
{
	if (ioptr > iobuf) {
		write(1, iobuf, ioptr - iobuf);
		ioptr = iobuf;
	}
}

static inline void echo_c(int c)
{
	if (ioptr >= ioend)
		ioflush();
	*ioptr++ = c;
}

static void echo(char *c)
{
	while (*c)
		echo_c((byte) *c++);
}

static void html_print(void)
{
	int x, y, d;
	int help = 1;
	
	echo("<table>");
	for (y = 0; y < Y; y++){
		echo("<tr>\n");
		for (x = 0; x < X; x++){
			d = maze[y][x];
			echo("<td ");
			if (! (d & (1 << 5))){
				if (d & (1 << 4)){
					echo("class='ex'>");
					echo("!");
					help = 0;			
				} else {
					echo("class='no'>");
					echo("?");
				}
			} else {
				echo("class='");
				if (d & 1){
					echo("up");
					if (d & (1 << 6))
						echo("r");
				}
				if (d & 2 ) {
					if (d & 1)
						echo(" ");
					echo("lf");
					if (d & (1 << 7))
						echo("r");
				}
				if (d & 4){
					if (d & 3)
						echo(" ");
					echo("dn");
					if (d & (1 << 8))
						echo("r");
				}
				if (d & 8){
					if (d & 7)
						echo(" ");
					echo("rt");
					if (d & (1 << 9))
						echo("r");
				}
				echo("'>");
				if (end_x == x && end_y == y) {
					echo("<img src=\"robot_karel.");
					echo_c("0132"[end_dir]);
					echo(".png\" width=16 height=16>");
				}
			}
		}
		echo("</tr>");
	}
	echo("</table>\n");
	if (help){
		echo(
"<p>Seznam příkazů:<br>\n"
"<b>jestart<br>\n"
"jezed<br>\n"
"konec<br>\n"
"krok<br>\n"
"nenistart<br>\n"
"nenized<br>\n"
"vlevo<br>\n"
"vpravo<br>\n"
"znovu</b>\n");
	}
}

static inline int ishex(int c)
{
	c |= 0x20;
	return (c >= '0' && c <= '9' || c >= 'a' && c <= 'f');
}

static inline int hexval(int c)
{
	c |= 0x20;
	c -= '0';
	if (c >= 10)
		c -= 0x27;
	return c;
}

static void cgi_read(void)
{
	char *method = getenv("REQUEST_METHOD");
	if (!method || strcasecmp(method, "POST"))
		return;
	stdin_read();

	byte *s = source;
	int c;
again:
	while (*s) {
		byte *line = s;
		while ((c = *s++) != '=') {
			if (!c || c == '\r' || c == '\n' || c == '&')
				goto again;
		}
		if (s-line == 8 && !memcmp(line, "program", 7)) {
			byte *d = source;
			while (*s && *s != '\r' && *s != '\n' && *s != '&') {
				int c = *s++;
				if (c == '+')
					*d++ = ' ';
				else if (c == '%' && ishex(s[0]) && ishex(s[1])) {
					*d++ = (hexval(s[0]) << 4) | hexval(s[1]);
					s += 2;
				} else
					*d++ = c;
			}
			*d = 0;
			source_len = d - source;
			return;
		} else {
			while (*s && *s != '\r' && *s != '\n' && *s != '&')
				s++;
			if (*s)
				s++;
		}
	}
	source_len = 0;
}

static void html_head(void)
{
	echo(
"Content-type: text/html; charset=utf-8\n"
"\n"
"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html40/strict.dtd\">\n"
"<html>\n"
"<head>\n"
"<title>Virtuální mašina</title>\n"
"<style type='text/css'>\n"
"table { border-collapse: collapse; }\n"
"td { width: 20px; height: 20px; border: 1px solid #ddd; text-align: center; vertical-align: middle; font-family: sans-serif; }\n"
"td.no { background-color: #aaa; color: #fff; font-weight: bold; }\n"
"td.ex { background-color: #00f; color: #fff; font-weight: bold; }\n"
"td.up { border-top: 2px solid black; }\n"
"td.dn { border-bottom: 2px solid black; }\n"
"td.lf { border-left: 2px solid black; }\n"
"td.rt { border-right: 2px solid black; }\n"
"td.upr { border-top: 2px solid red; }\n"
"td.dnr { border-bottom: 2px solid red; }\n"
"td.lfr { border-left: 2px solid red; }\n"
"td.rtr { border-right: 2px solid red; }\n"
"</style>\n"
"</head>\n"
"<body>\n"
"<h2>Úloha 10: Virtuální mašina</h2>\n"
);
}

static void html_tail(void)
{
	echo(
"<p>\n"
"<form method='post' action='.'>\n"
"<textarea cols=80 rows=25 name='program'>"
);
	for (int i=0; i<source_len; i++) {
		int c = source[i];
		if (c == '&')
			echo("&amp;");
		else if (c == '<')
			echo("&lt;");
		else if (c == '>')
			echo("&gt;");
		else
			echo_c(c);
	}
	echo(
"</textarea><br>\n"
"<input type='submit' value='Spustit'>\n"
"</form>\n"
"</body>\n"
"</html>\n"
);
	ioflush();
}

static void html_intro(void)
{
	echo(
"<ul>\n"
"<li>Zde je možno programovat náš virtuální stroj. Používá specifický\n"
"programovací jazyk, o kterém je známo jen to, že pro zápis programu stačí\n"
"malá písmena anglické abecedy a bílé znaky. Stroj ovšem akceptuje libovolný\n"
"vstup; znaky a příkazy, kterým nerozumí, jednoduše ignoruje.</li>\n"
"<li>Po kliknutí na tlačítko Spustit je zadaný program spuštěn ve virtuálním\n"
"stroji. Po doběhnutí programu je zobrazen výsledek jeho běhu.</li>\n"
"<li>Pokud program běží příliš dlouho, je zabit a není zobrazen jeho\n"
"výstup.</li>\n"
"</ul>\n"
"<p>\n"
);
}

int main(void)
{
	if (getenv("GATEWAY_INTERFACE")) {
		cgi_mode = 1;
		cgi_read();
		if (source_len) {
			parse();
			go();
			html_head();
			html_print();
			html_tail();
		} else {
			html_head();
			html_intro();
			html_tail();
		}
	} else {
		stdin_read();
		parse();
		go();
		dump();
	}
	return 0;
}
